"""SLI contains the logic needed to computed SLI/SLO metrics.

Calculate service level objective measurements from metrics stored
in common backends in accordance with the logic set out in the 
[SRE Workbook](https://landing.google.com/sre/workbook/toc/)

    Typical usage example:

    sli = Sli(metric_client)

    sli.window_length = 30  # days

    sli.slo = 0.99

    sli.get_metric_data()

    # Calculate sli/slo doing no group bys.

    sli.calculate()

    sli.error_budget()

    print(sli.slo_data)
"""

import time
from datetime import datetime
from google.cloud import monitoring_v3
import pandas as pd
from .metric_client import MetricClient

MetricDescriptor = monitoring_v3.enums.MetricDescriptor


class SliException():
    """Class to hold exceptions related to the SLI class specifically
    """

    class ValueNotSet(Exception):
        """Value Not Set exception
        Used when a class method is called without providing a value for a required class attribute
        """


    class UnsupportedMetricType(Exception):
        """Unsupported Metric Type exception
        Used if the selected metric type is not supported yet by the module
        """


class Sli():
    """Sli object for calculating sli/slo data

    Attributes:
        metric_client:      An instance of MetricClient object that will be used
                            to retrieve timeseries data.
        metric_data:        dataframe containing timeseries metric data. Either
                            provided manually or by the get_metric_data method
        window_end:         End of the window used to retrieve timeseries data
                            and calculate the sli. As seconds from the epoch,
                            eg time.time()
        window_length:      number of days over which to calculate the sli
        slo:                The service level objective e.g. 0.999
        group_by_labels:    metric labels by which to group sli calculation
    """

    def __init__(self, metric_client=MetricClient()):
        self.metric_client = metric_client
        self.metric_data = None
        self.window_end = time.time()
        self.window_length = 0
        self.slo = None
        self.group_by_labels = None

    @staticmethod
    def days_to_seconds(days):
        """Convert days into seconds

        Args:
            days: number of days

        Returns:
            The corresponding number of seconds
        """
        return days*24*60*60

    @property
    def window_length_seconds(self):
        """Access the window_length in seconds

        Returns:
            The window length, stored in days, as seconds
        """
        return Sli.days_to_seconds(self.window_length)

    @property
    def window_start(self):
        """Derive window start from end and duration

        Returns:
            The start of the window in seconds from epoch
        """
        return self.window_end - self.window_length_seconds

    def calculate(self):
        """
        Calculate SLI based on metric type
        Only boolean is supported right now
        """
        if self.metric_client.value_type == MetricDescriptor.ValueType.BOOL:
            self.calc_bool()
            self.add_period()
            self.add_slo()
            return self.slo_data
        elif self.metric_client.value_type == 'Some other type':
            # TODO
            pass
        else:
            raise SliException.UnsupportedMetricType

    def get_metric_data(self):
        """
        Retrieve metric data from the metric_client

        """
        if self.window_length is None:
            raise SliException.ValueNotSet("window_length cannot be None")
        self.metric_data = self.metric_client.timeseries_dataframe(
            end=self.window_end, duration=self.window_length_seconds
            )

    def get_bool_data(self):
        resampler = self.metric_data.resample(
            '30T', base=0, on='start_timestamp'
            )
        self.good = resampler.sum()
        self.valid = resampler.count()[['value']]

    def calc_bool(self):
        """

        """
        # self.get_metric_data()
        if self.group_by_labels:
            return self.calc_bool_agg()
        else:
            return self.calc_bool_simple()

    def calc_bool_agg(self):
        """
        Calculate sli aggregating over the columns in group_by_labels
        """
        good_events = self.metric_data.groupby(
            self.group_by_labels
            ).sum().reset_index().rename(columns={'value': 'count_good'})

        valid_events = self.metric_data.groupby(
            self.group_by_labels
            ).count().reset_index().rename(
                columns={'value': 'count_valid'}
                )['count_valid']

        slo_data = good_events.merge(
            valid_events,
            how='left',
            left_index=True,
            right_index=True)

        slo_data['sli'] = slo_data['count_good']/slo_data['count_valid']

        self.slo_data = slo_data
        return self.slo_data

    def calc_bool_simple(self):
        """
        Calculate sli over the entire dataframe with no aggregation by labels
        """
        valid_events = self.metric_data.shape[0]
        good_events = self.metric_data['value'].sum()
        self.slo_data = pd.DataFrame([
            {
                'count_good': good_events,
                'count_valid': valid_events,
                'sli': good_events/valid_events
            }
        ]

        )
        return self.slo_data

    def error_budget(self):
        """
        Using previously calculated good event and valid event count,
        return the error budget, and budget remaining.
        """
        if not self.slo:
            raise SliException.ValueNotSet("slo has not been defined")

        data = self.slo_data
        data['error_budget'] = \
            data['count_valid'] * (1-self.slo)
        data['error_budget_remaining'] = \
            data['error_budget'] - (data['count_valid'] - data['count_good'])
        self.slo_data = data
        return data

    def add_period(self):
        self.slo_data['period_from'] = datetime.fromtimestamp(
            self.window_end - self.window_length_seconds)
        self.slo_data['period_to'] = datetime.fromtimestamp(self.window_end)

    def add_slo(self):
        self.slo_data['slo'] = self.slo

    def add_labels(self):
        pass



"""pyslo

Module to help standardise the calculation of SLO data from metrics.

This module implements the logic defined in the
[SRE Workbook](https://landing.google.com/sre/workbook/toc/) in an
agnostic way.

In order to have SLO's reported consistently regardless of backend
timeseries database, aggregations and computations are all performed
within the module, as opposed to relying upon the backend's API
functionality.

Whilst this may not be the most efficient methodology if you know you
can exist in a single cloud environment, or utilize a single monitoring
framework, it is helpful if you are strung out over multiple clouds
and on prem environments.
"""

__version__ = '0.0.4'



"""Metric Client

Metric clients are used to retrieve time series data from a vairety of
sources. The MetricClient class can be inherited by provider specific derivatives.

Currently supported:
    - StackDriver
        Client library is in alpha. Using version v3:-
            https://googleapis.dev/python/monitoring/latest/gapic/v3/api.html

Planned implementations:
    - Azure Metric Service
    - Prometheus
"""

import time
import datetime
import pytz
import pandas as pd
from google.cloud import monitoring_v3

MetricDescriptor = monitoring_v3.enums.MetricDescriptor


class NoMetricDataAvailable(Exception):
    """ NoMetricDataAvailable

    Thrown if you try to retrieve data using invalid
    or out of range parameters
    """


class MetricClient():
    """Parent object for source specific metric clients.

    Attributes:
        project:       id of the GCP project hosting Stackdriver

    """

    value_type = None

    def timeseries_dataframe(self):
        """Retrieve data from time series db and return as a pandas dataframe
        """
        return


class StackdriverMetricClient(MetricClient):
    """Stackdriver Metric Client

    This client will allow you to retrieve metrics from a Stackdriver
    monitoring instance.

    Attributes:
        project:        string. id of the GCP project hosting Stackdriver
        metric_type:    string. the metric type as defined in stackdriver.
                        e.g. composer.googleapis.com/environment/healthy
                        This is most easily found by locating the metric in StackDriver Metrics
                        Explorer, then viewing as JSON.
                        The metric type is found as part of the timeSeriesFilter.
        value_type:     metric type, as a type defined in google.cloud.monitoring_v3
                        e.g. monitoring_v3.enums.MetricDescriptor.ValueType.BOOL

    """

    def __init__(self, project):
        self.project = project
        self.metric_type = None
        self.value_type = None

        self._client = monitoring_v3.MetricServiceClient()

    def timeseries_dataframe(self, end=time.time(), end_nanos=0, duration=3600):
        """Fetches and returns a dataframe of timeseries data

        By default this will retrieve the last hours worth of data.
        By specifying duration you will get that amount of data from now back.
        If you want a custom range then specify both end and duration.

        Args:
            end:        Optional. A float that represents the end of the period, as the time in
                        seconds since the epoch. default is now. Value will be converted to an
                        integer second. Use end_nanos if you need nanosecond precision.
            end_nanos:  Optional. Integer number of nano seconds that will be added to the end
                        value. deafult = 0
            duration:   Optional. An integer length of the period to retrieve in seconds.
                        default = 3600s

        Returns:
            A pandas dataframe
        """
        interval = self.set_interval(end, end_nanos, start_time=(end - duration))
        iterator = self.get_timeseries_iter(interval)
        materialized = self.fetch_iter_results(iterator)
        return self.to_df(materialized)

    @property
    def filter(self):
        """Formats a metric filter

        Returns:
            A string that can be passed as the filter_ arg during a MetricServiceClient
            list_time_series call.  The filter specifies which time series data is being requested.
            The filter must specify a single metric type, and can additionally
            specify metric labels and other information. For example:

            metric.type = "compute.googleapis.com/instance/cpu/usage_time" AND
                metric.labels.instance_name = "my-instance-name"

        Todo: - for now we are just making the simplest case filter!
        """
        return f'metric.type = "{self.metric_type}"'

    def get_timeseries_iter(self, interval):
        """Retrieves timeseries data from Stackdriver

        Args:
            interval: google.cloud.monitoring_v3.types.TimeInterval()

        Returns:
            A page or results iterator.
        """
        project_name = self._client.project_path(self.project)
        results_iter = self._client.list_time_series(
            project_name,
            self.filter,
            interval,
            monitoring_v3.enums.ListTimeSeriesRequest.TimeSeriesView.FULL
        )
        return results_iter

    @staticmethod
    def get_labels(result):
        """Extract the resource labels from the result object

        Args:
            result:

        Returns:
            a 'google.protobuf.pyext._message.ScalarMapContainer'. Basically
            a dict type object with the labels.
        """
        return result.resource.labels

    def to_df(self, results):
        """
        For a set of results, create a dataframe.
        Dastaframe will include the result resource labels as
        columns.
        """
        points = list()
        for result in results:
            labels = self.get_labels(result)
            print(type(labels))
            for point in result.points:
                start_time = self.convert_point_time(
                    point.interval.start_time,
                    as_timestamp=False
                    )

                end_time = self.convert_point_time(
                    point.interval.end_time,
                    as_timestamp=False
                )

                point_dict = {
                        'start_timestamp': start_time,
                        'end_timestamp': end_time,
                        'value': self.get_point_value(point.value)
                        }
                point_dict.update(labels)
                points.append(point_dict)
        if len(points) == 0:
            raise NoMetricDataAvailable

        return pd.DataFrame(points)

    def get_point_value(self, point_value):
        """EXtract value from point_value object

        Args:
            point_value: instance of google.cloud.monitoring_v3.types.TypedValue
        Returns:
            The value of the TypedValue object
        """
        if self.value_type == MetricDescriptor.ValueType.BOOL:
            return int(point_value.bool_value)
        else:
            raise Exception

    @staticmethod
    def convert_point_time(point_time, as_timestamp):
        """Convert a TypedValues time to a datetime value
        
        Args:
            point_time: google.cloud.monitoring_v3.types.TypedValue

        Returns:
            datetime object. Note datetime object only supports microsecond
            accuracy.
        """
        seconds = point_time.seconds
        nanos = point_time.nanos
        if as_timestamp:
            return seconds + nanos/10**9
        else:
            return datetime.datetime.fromtimestamp(seconds + nanos/10**9, tz=pytz.UTC)


    @staticmethod
    def fetch_iter_results(results_iterator):
        results = list()
        for result in results_iterator:
            results.append(result)
        return results


    @staticmethod
    def set_interval(end_time, end_time_nanos=0, start_time=None):
        """Create a TimeInterval object based on input start and end times

        For GAUGE kind metrics, a time interval could be provided to the API with no start_time
        specified. In this case api client assumes start=end and will try to retrieve a single
        data point. In this application there is not much use for that, but its supported in case
        you are trying to retrieve a single GAUGE metric point.
        We support setting the end time nanoseconds, which is unlikely to be needed in an SLO
        calculation but we do not bother to set the start_nanos.

        Args:
            end_time:       the end of the period, as the time in seconds since the epoch.
                            If a float is provided it is converted to an int regardless
                            (which will round the float down).
            end_time_nanos: Optional. Provide the number of nano seconds that will be added
                            to the end_time.
            start_time:     the start of the period in seconds since the epoch.

        Returns:
            An instance of google.cloud.monitoring_v3.types.TimeInterval()
        """
        interval = monitoring_v3.types.TimeInterval()  # pylint: disable=no-member
        interval.end_time.seconds = int(end_time)
        interval.end_time.nanos = int(end_time_nanos)
        if start_time:
            interval.start_time.seconds = int(start_time)
        return interval
